// Ethernet (Wiznet W5500) specific routines for USB+Ethernet/DMX Generator **
//
// Simple DMX512(incl. Art-Net) controller for STM32F042F6 (20pin TSSOP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.c", "main.h", "usb-fs.c", "dmx.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics and EMBITZ.org
//
// Contact: www.midi-and-more.de
// A detailled hardware construction and operation manual
// is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz 1.0  does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "globaldefines.h"

//from main
extern struct PARAM{
uint32_t VidPid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t Gateway;
uint32_t ArtnetIP;
uint32_t ArtSubnet;
uint32_t ArtUniverse;
uint32_t Channel;
} gComParam;  //10 words

extern uint16_t gGREEN_LEDCOUNT, gTIMEOUT;
extern void servSystem(void);

extern void msDelay(uint16_t) ;
extern void sendUsbB(uint8_t);
extern void sendUsbX16(uint16_t);
extern void sendUsbX32(uint32_t);

//from dmx.c
extern uint8_t gdDMXBUF[DMX_BUFLEN+7];          //actual DMX transmit buffer
extern uint8_t gdFINALMEM[DMX_BUFLEN+7];        //final level of fade process

// ************************************

uint8_t geARTDMXSTR[14] = {'A','r','t','-','N','e','t',0,0,0x50,0,14,0,0} ;
uint8_t geSHORTNAME[18] = {'U','S','B','/','D','M','X','+','A','r','t',' ','G','e','n',0,0,0};
uint8_t geLONGNAME[24] = {'U','S','B','/','D','M','X','+','A','r','t','n','e','t',' ',
                          'G','e','n','e','r','a','t','o','r'};
uint8_t geREPORTSTR[13] = {'#','0','0','0','1','[','0','0','0','0',']','o','k'};

/*
.DB "Art-Net",0,0,0x50,0,14  ;12 Bytes
		;Byte 0..7 = Artnet ID
		;Byte 8 = Opcode Low
		;byte 9 = Opcode High ("OpOutput" = 0x5000)
		;Byte10 = ProtVerH = 0
		;Byte11 = ProtVerL = 14
		;Byte12 = Sequence 0=disabled
		;Byte13 = Physical (I/O num.)
*/

#define SPI_BUFLEN 700
//SPI buffers are linear, not cyclic!!
uint8_t geSPI_RXBUF[SPI_BUFLEN];
uint8_t *geSPI_RXHEAD=0, *geSPI_RXTAIL=0, *geSPI_RXBUFTOP= &(geSPI_RXBUF[SPI_BUFLEN]);
uint8_t geSPI_TXBUF[SPI_BUFLEN/2];

uint8_t geDummy;            //global only to avoid warning during compilation
uint8_t geArtnetMsg;        //special Art-Net message types
uint8_t geArtnetOn;         //1=default: Artnet data are send to DMX. 0=block

uint8_t initW5500(void);
uint8_t openUDP0(void);
uint8_t addEtherQuery(uint8_t);
uint8_t checkEtherSend(void);
uint8_t sendEtherQuery(void);
uint16_t sendEtherBuffer(uint8_t *, uint16_t);
uint8_t receiveUDP(void);
void artPollReply(void);

void initSPI1(void);
void wait1us(void);
void sendSpiWord(uint16_t);
uint16_t receiveSpiWord(void);
uint16_t exchangeSpiWord(uint16_t);
uint8_t exchangeSpiByte(uint8_t);
void writeWiz_4(uint16_t, uint8_t, uint8_t);
uint8_t readWiz_4(uint16_t, uint8_t);
void writeWiz_5(uint16_t, uint8_t, uint16_t);
uint16_t readWiz_5(uint16_t, uint8_t);
void writeWiz_7(uint16_t, uint8_t, uint32_t);
uint32_t readWiz_7(uint16_t, uint8_t);
uint16_t writeWiz_Data(uint16_t, uint8_t*, uint16_t);
uint16_t readWiz_Data(uint16_t, uint16_t);

/*
//W5500 Common Registers (SPI Adressing)
//--------------------------------------
#define COMMON_BLOCK    0x00
#define MR              0x0000  //Mode Register, for S/W Reset
#define GAR0            0x0001  //4 Gateway Address Register
#define SUBR0           0x0005  //4 Subnet Mask Register
#define SHAR0           0x0009  //6 Source Hardware Address Register
#define SIPR0           0x000F  //4 Source IP Address Register
#define IR              0x0015  //Interrupt Register
#define IMR             0x0016  //Interrupt Mask Register
#define SIR             0x0017  //Socket Interrupt Register
#define SIMR            0x0018  //Socket Interrupt Mask Register
#define RTR0            0x0019  //Word Initial Retry Time-value Register
#define RCR             0x001B  //Retry Count Register
#define PHYCFGR         0x002E  //PHY Configuration Register

//W5500 Socket0 Registers
//-----------------------
#define S0_REGBLOCK     0x01
#define S0_MR           0x0000  //Socket 0 Mode Register
#define S0_CR           0x0001  //Socket 0 Command Register
#define S0_IR           0x0002  //Socket 0 Interrupt Register
#define S0_SR           0x0003  //Socket 0 Status Register
#define S0_PORT0        0x0004  //Socket 0 Local Port Register
#define S0_DHAR0        0x0006  //Socket 0 Destination Hardware Address
#define S0_DIPR0        0x000C  //Socket 0 Destination IP Address
#define S0_DPORT0       0x0010  //Socket 0 Destination Port Register
#define S0_MSSR0        0x0012  //Socket 0 Maximum Segment Size
#define S0_TTL			0x0016  //Socket 0 Time to Live Register
#define S0_RXBUF_SIZE   0x001E  //Socket 0 Receive Buffer Size
#define S0_TXBUF_SIZE   0x001F  //Socket 0 Transmit Buffer Size
#define S0_TX_FSR0      0x0020  //Socket 0 TX Free Size
#define S0_TX_RD0       0x0022  //Socket 0 TX Read Pointer
#define S0_TX_WR0       0x0024  //Socket 0 TX Write Pointer
#define S0_RX_RSR0      0x0026  //Socket 0 RX Received Size
#define S0_RX_RD0       0x0028  //Socket 0 RX Read Pointer
#define S0_RX_WR0       0x002A  //Socket 0 RX Write Pointer
#define S0_IMR          0x002C  //Socket 0 Interrupt Mask Register
//#define S0_KPALVTR      0x002F  //Socket 0 Keep Alive Timer (TCP only)

#define S0_TXBLOCK      0x02
#define S0_RXBLOCK      0x03
*/


// ########################################################################

uint8_t initW5500(void)   // **********************************************
// initializes Wiz5500 registers for desired mode of operation.
// **************************************************************************
{
uint32_t artSubnetMask;

//PortC,4 handles W5500 /RST
    GPIOB->BSRR = GPIO_BSRR_BR_1;                               //set /RST low
    msDelay(5);
    GPIOB->BSRR = GPIO_BSRR_BS_1;                               //set /RST high
    msDelay(100);                                               //min 50ms

    if (gComParam.ArtnetIP == 0) artSubnetMask = gComParam.Subnet; //overwrite
    else artSubnetMask = 0xFF000000;

//allocate maximum socket memory (16kB)
    writeWiz_4(S0_RXBUF_SIZE,S0_REGBLOCK,16);
    writeWiz_4(S0_TXBUF_SIZE,S0_REGBLOCK,16);
//transfer MAC ID from gComParam
    writeWiz_5(SHAR0,COMMON_BLOCK,(uint16_t)(gComParam.MacHi & 0x0000FFFF));
    writeWiz_7(SHAR0+2,COMMON_BLOCK,gComParam.MacLo);
//transfer Gateway IP from gComParam
    writeWiz_7(GAR0,COMMON_BLOCK,gComParam.Gateway);
//transfer Subnet Mask from gComParam
    writeWiz_7(SUBR0,COMMON_BLOCK,artSubnetMask);
//transfer Local IP from gComParam
    writeWiz_7(SIPR0,COMMON_BLOCK,gComParam.LocalIP);

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear pending interrupts
    return 1;
}


uint8_t openUDP0(void)   // *************************************************
// with entries from gComParam
// **************************************************************************
{
uint8_t recvbyte;
uint32_t artDestIP, artIpB;

    if(gComParam.ArtnetIP==0) artDestIP=gComParam.LocalIP | ~(gComParam.Subnet);
    else  {
        artIpB = (((gComParam.MacLo>>16) + OEMH + OEML) & 0xFF) << 16;
        sendUsbX32(artIpB);
        artDestIP =(gComParam.ArtnetIP<<24)|artIpB|(gComParam.MacLo&0x0000FFFF);
    }

//transfer UDP Local Port from gComParam
   writeWiz_5(S0_PORT0,S0_REGBLOCK,ARTNETPORT);
//transfer UDP Destination Port from gComParam
    writeWiz_5(S0_DPORT0,S0_REGBLOCK,ARTNETPORT);
//transfer UDP Destination IP from gComParam
    writeWiz_7(S0_DIPR0,S0_REGBLOCK,artDestIP);
//activate UDP Unicast or Broadcast mode
    writeWiz_4(S0_MR,S0_REGBLOCK,0x02);                         //std.UDP mode
    writeWiz_4(S0_CR,S0_REGBLOCK,0x01);                         //=CMD_OPEN
    recvbyte = 0;
    while ((recvbyte= readWiz_4(S0_SR,S0_REGBLOCK)) != 0x22 );  //=SOCK_UDP

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear interrupt flags
    writeWiz_4(S0_IMR,S0_REGBLOCK,0);               //no interrupts
    writeWiz_4(SIR,COMMON_BLOCK,0);
    writeWiz_4(SIMR,COMMON_BLOCK,0);

    return 1;
}


uint16_t sendEtherBuffer(uint8_t *buffer, uint16_t toSend)  // **************
// demands linear buffer, not cyclic !
// **************************************************************************
{
uint16_t temp, txPtr;

    temp = readWiz_4(S0_SR,S0_REGBLOCK);
    temp = readWiz_5(S0_TX_FSR0,S0_REGBLOCK);

    if (temp < toSend) return 0;                    //no buffer available

    txPtr = readWiz_5(S0_TX_WR0,S0_REGBLOCK);
    txPtr = writeWiz_Data(txPtr, buffer, toSend);
    writeWiz_5(S0_TX_WR0,S0_REGBLOCK,txPtr);        //update W5500 write pointer
    writeWiz_4(S0_CR,S0_REGBLOCK,0x20);             //=CMD_SEND

    return 1;
}


uint8_t receiveUDP(void)  // ************************************************
//if available, one UDP packet is transferred
//from W5500 internal receive buffer to geSPI_RXBUF (linear,not cyclic!!)
//buffer pointers are updated
//Art-Net special information is evaluated
// **************************************************************************
{

uint16_t rxCount, rxPtr, udpLength, k;
uint32_t udpHeader;

    rxCount = readWiz_5(S0_RX_RSR0,S0_REGBLOCK);
    if (rxCount <= 8) return 0; //no data received, empty UDP packet or error

    rxPtr = readWiz_5(S0_RX_RD0,S0_REGBLOCK);
    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxPtr += 4;
    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxCount -= 8;
    rxPtr += 4;
    udpLength =(uint16_t)(udpHeader & 0x0000FFFF);
                                        //what if geUdpLength != rxPtr ????????
    geSPI_RXHEAD = geSPI_RXBUF;
    geSPI_RXTAIL = geSPI_RXBUF;
    rxPtr = readWiz_Data(rxPtr,udpLength);      //copy message into geSPI_RXBUF
    writeWiz_5(S0_RX_RD0,S0_REGBLOCK,rxPtr);    //update W5500 read pointer
    writeWiz_4(S0_CR,S0_REGBLOCK,0x40);         //=CMD_RECV

//process Art-Net message
    geArtnetMsg = geSPI_RXBUF[9];
    if (geArtnetMsg == 0x20) {
        artPollReply();
        return 1;
    }
    else if (geArtnetMsg != 0x50) return 0;
//handle Art-Net data packet:
//redefine udpLength
    udpLength = ((uint16_t)geSPI_RXBUF[16] << 8) + (uint16_t)geSPI_RXBUF[17];
    if (geArtnetOn)  {                      //0 = block Artnet, only USB control
        for (k=0;k<udpLength;k++)  {
            gdDMXBUF[k+1] = geSPI_RXBUF[k+18];
            gdFINALMEM[k+1] = geSPI_RXBUF[k+18];
        }
        GPIOB->BSRR = GPIO_BSRR_BR_8;       // clear bit PC3, green LED OFF
        gGREEN_LEDCOUNT = 200;              //milliseconds
    }
    return 1;
}


void artPollReply(void)   // ************************************************
// **************************************************************************
{
uint16_t k;
uint8_t n;
uint32_t temp;

    k = 0;
    n = 0;
    for (n=0;n<9;n++) {
        geSPI_TXBUF[k++] = geARTDMXSTR[n];
    }
    geSPI_TXBUF[k++] = 0x21;        //Opcode "ArtPollReply"
    temp = gComParam.ArtnetIP;
    n = (temp>>24) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (temp>>16) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (temp>>8) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = temp & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(ARTNETPORT & 0x00FF);
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((ARTNETPORT >> 8) & 0x00FF);
    geSPI_TXBUF[k++] = n;
    geSPI_TXBUF[k++] = HARDWARE_VERSION;    //byte 16
    geSPI_TXBUF[k++] = REVNUM;
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = (uint8_t)(gComParam.ArtSubnet & 0x0000000F);
    geSPI_TXBUF[k++] = OEMH;
    geSPI_TXBUF[k++] = OEML;
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0xD0;    //Node Status= Indicators normal
                                //Front Panel setup, kein Dual Boot
    geSPI_TXBUF[k++] = 0x69;    //'i' Cinetix ESTA Low  byte 24
    geSPI_TXBUF[k++] = 0x43;    //'C' Cinetix ESTA High
//REMOTESHORT einplanen
    for (n=0;n<18;n++) {
        geSPI_TXBUF[k++] = geSHORTNAME[n];  //bytes 26 to 43
    }
//REMOTELONG einplanen
    for (n=0;n<24;n++) {
        geSPI_TXBUF[k++] = geLONGNAME[n];   //bytes 44-107 (total 64char)
    }
        for (n=0;n<40;n++) {
        geSPI_TXBUF[k++] = 0;               //filling bytes of LONGNAME
    }
    for (n=0;n<13;n++) {
        geSPI_TXBUF[k++] = geREPORTSTR[n];   //bytes 108-171 (total 64char)
    }
        for (n=0;n<51;n++) {
        geSPI_TXBUF[k++] = 0;               //filling bytes of REPORTSTR
    }
    geSPI_TXBUF[k++] = 0;       //NumPortsH = byte 172
    geSPI_TXBUF[k++] = 1;       //NumPortsL = byte 173

    geSPI_TXBUF[k++] = 0x80;    //PortTypes: one DMX512 output byte 174
    for (n=0;n<7;n++) {         //all out-commented fills concatentated
        geSPI_TXBUF[k++] = 0;   //
    }
//compacted to reduce size of source code:
//    geSPI_TXBUF[k++] = 0;       //fills bytes 175-177
//    geSPI_TXBUF[k++] = 0;
//    geSPI_TXBUF[k++] = 0;

//    geSPI_TXBUF[k++] = 0;       //Good Input =0 bytes 178-181
//    geSPI_TXBUF[k++] = 0;       //fills
//    geSPI_TXBUF[k++] = 0;
//    geSPI_TXBUF[k++] = 0;

    geSPI_TXBUF[k++] = 0x80;    //Good Output  bytes 182-185
    for (n=0;n<7;n++) {         //all out-commented fills concatentated
        geSPI_TXBUF[k++] = 0;   //
   }
//    geSPI_TXBUF[k++] = 0;       //fills
//    geSPI_TXBUF[k++] = 0;
//    geSPI_TXBUF[k++] = 0;

//    geSPI_TXBUF[k++] = 0;       //Swin: dummy:no Rx Uinverse bytes 186-189
//    geSPI_TXBUF[k++] = 0;       //fills
//    geSPI_TXBUF[k++] = 0;
//    geSPI_TXBUF[k++] = 0;

    n= (gComParam.ArtSubnet & 0x0000000F)<<4;
    n |= gComParam.ArtUniverse & 0x0000000F;
    geSPI_TXBUF[k++] = n;       //Swout  bytes 190 - 193
//    geSPI_TXBUF[k++] = 0;       //fills
//    geSPI_TXBUF[k++] = 0;
//    geSPI_TXBUF[k++] = 0;

    for (n=0;n<10;n++) {        //all fills concatentated
        geSPI_TXBUF[k++] = 0;   //194-199: div. special features = 0
    }
//    geSPI_TXBUF[k++] = 0;       //200:Style: "a DMX to / from Art-Net device"
    n = (uint8_t)((gComParam.MacHi >> 8) & 0x000000FF);  //201
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(gComParam.MacHi& 0x000000FF);  //202
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 24) & 0x000000FF);  //203
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 16) & 0x000000FF);  //204
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 8) & 0x000000FF);  //205
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(gComParam.MacLo & 0x000000FF);  //206
    geSPI_TXBUF[k++] = n;
    for (n=0;n<32;n++) {
        geSPI_TXBUF[k++] = 0;   //207-238
    }

    sendEtherBuffer(geSPI_TXBUF,238);

    GPIOB->BSRR = GPIO_BSRR_BR_8;   // clear bit PB8, green LED OFF
    gGREEN_LEDCOUNT = 200;          //milliseconds
}


// ##########################################################################
// low level routines

void initSPI1(void)   // ****************************************************
//SPI1 is Master
// **************************************************************************
{

     RCC->APB2ENR |=  RCC_APB2ENR_SPI1EN ;   // SPI1 clock enable

//PA5 is SPI1_SCK
//PA6 is SPI1_MISO
//PA7 is SPI1-MOSI, configure as Alternate Function
    GPIOA->MODER&= ~(GPIO_MODER_MODER5 | GPIO_MODER_MODER6 | GPIO_MODER_MODER7);
    GPIOA->MODER|=(GPIO_MODER_MODER5_1|GPIO_MODER_MODER6_1|GPIO_MODER_MODER7_1);

// /INT is not connected (not used for this application)
//PA4 is /SCN output
    GPIOA->MODER &= ~(GPIO_MODER_MODER4);
    GPIOA->MODER |= (GPIO_MODER_MODER4_0);
    GPIOA->OTYPER &= ~(GPIO_ODR_4);         //PA4 push-pull
//set GPIOA 4,5,6,7 Speed up to 25MHz ("middle speed")
    GPIOA->OSPEEDR &= 0xFFFF00FF ;          //OSPEEDRx_1 down
    GPIOA->OSPEEDR |= 0x00005500 ;          //OSPEEDRx_0 up
//PB1 is /RST output
    GPIOB->MODER &= ~(GPIO_MODER_MODER1);
    GPIOB->MODER |= (GPIO_MODER_MODER1_0);
    GPIOB->OTYPER &= ~(GPIO_ODR_1);         //PB1 push-pull
//SPI1 is AF0
    GPIOA->AFR[0] &= 0xFFFFF000;            //configure PA5,6,7 AF function
    GPIOA->AFR[0] |= 0x00000000;            //formally

    geSPI_RXHEAD = geSPI_RXBUF;
    geSPI_RXTAIL = geSPI_RXBUF;

//SPI Master configuration
    SPI1->CR1 = SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_MSTR;
                                //idle=low,sample rising edge,freq=1/2APB2 clock
    SPI1->CR2 =  SPI_CR2_DS;    //default 16bit, polling

    GPIOB->BSRR = GPIO_BSRR_BS_1;                               //set /RST high
    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high

    SPI1->CR1 |= SPI_CR1_SPE;                                   //enable
}

/*
//was contained in previous firmware versions and worked acceptably
//removed, because critical and somewhat obscure together with STM32

uint8_t exchangeSpiByte(uint8_t data)   // **********************************
// **************************************************************************
{
uint8_t recv;

    SPI1->CR2 = SPI_CR2_FRXTH | 0x0700 ;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
//STM32 processors support 8 bit SPI transfer very badly, always 16 clock pulses
//for this reason, when data blocks with even count are transferred,
//exchangeSpiByte() must always be the last transfer command
//and /SCN must be set after 8bits are transferred
//if possible, avoid exchangeSpiByte()

    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = (uint16_t)data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    recv = SPI1->DR;

    wait1us();    //special hack because some STM32 SPI always exchange 16 bit
//???????? special hack for fine adjustment of pulse edges ?????
//seemed necessary for a while during firmware development,
//but now it works without. Exact reason unknown.

    while (SPI1->SR & SPI_SR_RXNE) {
        geDummy = SPI1->DR;
    }

    SPI1->CR2 =  SPI_CR2_DS;
    return recv;
}


void wait1us(void)   // ********************************************
//blocking delay by a little bit more than 1 microsecond
// **************************************************************************
{
uint32_t countdown;

	countdown = (uint32_t)SystemCoreClock/12000000;
    while (countdown)  countdown-- ;
}
*/

uint16_t exchangeSpiWord(uint16_t data)   // ********************************
// **************************************************************************
{
    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    return SPI1->DR;
}


void writeWiz_4(uint16_t address, uint8_t block, uint8_t data)   // *******
// write 4 byte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low

    exchangeSpiWord(address);
    send = ((block<<11) | 0x0400 | data);
    exchangeSpiWord(send) ;

    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
}


void writeWiz_5(uint16_t address, uint8_t block, uint16_t data) // **********
// write 5 byte Wiz5500 compatible cluster (1 halfword, e.g. port no.)
// **************************************************************************
{
uint16_t send;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
    send = (block<<11) | 0x0400 | (data >> 8) ;
    exchangeSpiWord(address);
    exchangeSpiWord(send) ;

  	GPIOA->BSRR = GPIO_BSRR_BS_4;       					    //set /SCN high
    send = (block<<11) | 0x0400 | (data & 0x00FF);
    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address+1);
    exchangeSpiWord(send) ;

  	GPIOA->BSRR = GPIO_BSRR_BS_4;       					    //set /SCN high
}


void writeWiz_7(uint16_t address, uint8_t block, uint32_t data)  // *********
// write 7 byte Wiz5500 compatible cluster (1 word, e.g. IP address)
// **************************************************************************
{
    writeWiz_5(address,block,(uint16_t)((data >> 16) & 0x0000FFFF));
    writeWiz_5(address+2,block,(uint16_t)(data & 0x0000FFFF));
}


uint8_t readWiz_4(uint16_t address, uint8_t block)   // *******************
// read 1 databyte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t recv, send;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11);
    recv = exchangeSpiWord(send) ;
    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high

    return ((uint8_t)(recv & 0x00FF));
}


uint16_t readWiz_5(uint16_t address, uint8_t block)   // ********************
// read 1 halfword Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send, recv0, recv1;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low

    send = (block<<11);
    exchangeSpiWord(address);
    recv0 = exchangeSpiWord(send);

    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    send = (block<<11);
    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
                                                // /SCN is high for about 0.5 us
    exchangeSpiWord(address+1);
    recv1 = exchangeSpiWord(send);

    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high

    return ((recv0 & 0x00FF) << 8) | (recv1 & 0x00FF);
}


uint32_t readWiz_7(uint16_t address, uint8_t block)   // ********************
// read 1 word Wiz5500 compatible cluster
// **************************************************************************
{
uint32_t  recv2, recv0;
    recv0 = (uint32_t)readWiz_5(address,block);
    recv2 = (uint32_t)readWiz_5(address+2,block);
    return ((recv0 << 16) | (recv2 & 0x0000FFFF));
}


uint16_t writeWiz_Data(uint16_t address, uint8_t* buffer, uint16_t toSend) // *
// write (toSend+3) byte Wiz5500 compatible cluster (i.e. toSend databytes)
// **************************************************************************
{
uint16_t send;
uint8_t byte1, byte2;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low

    if ((toSend & 1) == 0)  {               //an even number of bytes to be sent
    send = ((S0_TXBLOCK<<11)|0x0400 |*buffer++); //linear buffer, no wrap around
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;
    }

    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    send = ((S0_TXBLOCK<<11) | 0x0400 | *buffer++);

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;

//from here, only zero or an even number of bytes to be sent
    while (toSend)  {
        byte1 = *buffer++;
        byte2 = *buffer++;
        exchangeSpiWord(byte1<<8 | byte2);
        toSend -= 2;
        address += 2;
    }
    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    return address;
}


uint16_t readWiz_Data(uint16_t address,uint16_t toRead)  // *****************
// read (count) databytes from Wiz5500 Rx Buffer to geSPI_RXBUF
// **************************************************************************
{
uint16_t recvword, send;

    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low

    if ((toRead & 1) == 0)   {              //an even number of bytes to be read
        send = (S0_RXBLOCK<<11);
        exchangeSpiWord(address);
        recvword = exchangeSpiWord(send);
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead--;
        address++;
    }

    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    send = (S0_RXBLOCK<<11);
    send = (S0_RXBLOCK<<11);                //repeat for a little extra delay
    GPIOA->BSRR = GPIO_BSRR_BR_4;                               //set /SCN low
                                            // /SCN is high for about 0.5 us
    exchangeSpiWord(address);
    recvword = exchangeSpiWord(send);
    *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
    geSPI_RXHEAD++;
    if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
    toRead--;
    address++;

//from here, only zero or an even number of bytes to be read
    while (toRead)  {
        recvword = exchangeSpiWord(0);
        *geSPI_RXHEAD = (uint8_t)((recvword >> 8) & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead -= 2;
        address += 2;
    }
    GPIOA->BSRR = GPIO_BSRR_BS_4;                               //set /SCN high
    return address;
}

//###########################################################################

