// USB+Ethernet/DMX Generator  ***********************************************
//
// Simple DMX512(incl. Art-Net) controller for STM32F042F6 (20pin TSSOP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.h", "usb-fs.c", "dmx.c", "ether.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics and EMBITZ.org
//
// Contact: www.midi-and-more.de
// A detailled hardware construction and operation manual
// is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz 1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "usb_regs.h"
#include "globaldefines.h"
#include "main.h"


int main(void)    // ********************************************************
// **************************************************************************
{
uint16_t recvword;
uint8_t recvbyte, index, status, data1, data2;

    SCB->CCR |= SCB_CCR_STKALIGN_Msk; // Enable double word stack alignment
//(recommended in Cortex-M3 r1p1, default in Cortex-M3 r2px and Cortex-M4)

//effectively executed code from NVIC_PriorityGroupConfig(uint32_t NVIC_PriorityGroup) ::
#define AIRCR_VECTKEY_MASK    ((uint32_t)0x05FA0000)
#define NVIC_PriorityGroup_0  ((uint32_t)0x700)
//                              !< 0 bits for pre-emption priority
//                              4 bits for subpriority

// Set the PRIGROUP[10:8] bits according to NVIC_PriorityGroup value
    SCB->AIRCR = AIRCR_VECTKEY_MASK | (uint32_t)0x700;

    RCC->AHBENR |= RCC_AHBENR_GPIOAEN;      // Enable Port A clock
    RCC->AHBENR |= RCC_AHBENR_GPIOBEN;      // Enable Port B clock

// PA0 is used as red LED output
    GPIOA->MODER |= GPIO_MODER_MODER0_0 ; //PortA0 general purpose output mode
    GPIOA->BSRR = GPIO_BSRR_BS_0;         //Set bit PA0 (red LED ON)
// PB8 is used as green LED output
    GPIOB->MODER |= GPIO_MODER_MODER8_0 ; //PortB8 general purpose output mode
    GPIOB->BSRR = GPIO_BSRR_BS_8;         //Set bit PB8 (green LED ON)
    //GPIOx->OTYPER  is push-pull by default
    //GPIOx->OSPEEDR = 0;  //2 MHz by default
    //GPIOx->PUPDR  is HiZ by default

// PA3 enables/disables the DMX driver
    GPIOA->MODER |= GPIO_MODER_MODER3_0 ; //PortA3 general purpose output mode
    GPIOA->BSRR = GPIO_BSRR_BR_3;         //clear bit PA3 (DMX driver disabled)

// for USB, remap PA11,12 to PA9,10
    RCC->APB2ENR |= RCC_APB2ENR_SYSCFGCOMPEN;
    SYSCFG->CFGR1 |= SYSCFG_CFGR1_PA11_PA12_RMP;

    gSYSCOUNT=20;
    SysTick_Config(SystemCoreClock/1000);   //1ms / 1kHz SysTick

    gdMiniDmxOn = 0;
    gcIgnoreVelo0 = 0;
    gdFaderBusy = 0;
    gVidShadow = 0;             //as precaution, will be set finally
    gPidShadow = 0;             //by popFlashPage25()
    gGREEN_LEDCOUNT=0;
    gRED_LEDCOUNT=0;
    gDELAYCOUNT=0;
    gTIMEOUT=0;

#ifndef DMX_DEBUG
//start Independent Watchdog   *******************
//will be served by "pullUsartRxBuf" and "pullUsbRxBuf"
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog
#endif

//check jumper or front panel switch and set guDeviceClass  ******************
    GPIOA->PUPDR |=  GPIO_PUPDR_PUPDR14_0;      //does not conflict with SWDebug
    GPIOA->PUPDR &=  ~GPIO_PUPDR_PUPDR14_1;
    msDelay(2);
//when PortA,14 = high (default): CDC_CLASS
//when PortA,14 = low (Jumper set): MIDI_CLASS
    guDeviceClass = CDC_CLASS;
    geArtnetOn = 1;                 //default 1= don't block Artnet, 0=block
    if ((GPIOA->IDR & m14) == 0){
        guDeviceClass = MIDI_CLASS;
        geArtnetOn = 0;                     //may be unblocked manually
    }

//init I/O state  *********************
    popFlashPage17();           //copy user setup to SRAM for easier access
    popPresetPage(0);           //load user stored lighting scene to DMXBUF

// init communication **************************

    startUsb();

    initSPI1();
    initW5500();
    openUDP0();

#ifdef DMX_DEBUG
    initUsart2();
    msDelay(10);
    sendB2('$');
#else
// init DMX transmitter
    GPIOA->BSRR = GPIO_BSRR_BS_3;           //set bit PA3 (enable DMX driver)
    startDmxTx();                           //transmits even in Setup Mode
    DmxHeader();
#endif

 	gcPushByte = 0;

    while(1)
    {
        servSystem();                       //essential, controls DMX timing
        if (guDeviceClass == CDC_CLASS)  {
            if (gcPushByte)  {
                recvbyte = gcPushByte;      //uses next command code as
                gcPushByte = 0;             //flexible parameter input delimiter
                if (gcSetupMode) Setup_Interpreter(recvbyte);
                else  ASCII_Interpreter(recvbyte);
            }
            recvword = pullUsbRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                if (recvbyte == 'Z')  {
                    gTIMEOUT = 1000;
                    geArtnetOn = 0;         //block always
                    gdFaderBusy = 0;        //block fade process
                    MiniDMX_Interpreter();
                }
                else if (gcSetupMode) Setup_Interpreter(recvbyte);
                else  ASCII_Interpreter(recvbyte);
            }
            else if ((gTIMEOUT == 0) && gdMiniDmxOn) {
                gdMiniDmxOn = 0;            //terminate MiniDMX
                geArtnetOn = 1;             //re-activate Art-Net
            }
        }
        else {                              //MIDI_CLASS
            recvword = pullUsbRxBuf();
            if (recvword)  {
                index = (uint8_t)(recvword & 0x00FF);
                if (index > 0x0C) index = 0;

                recvword = pullUsbRxBuf();
                if (recvword)  status = (uint8_t)(recvword & 0x00FF);
                else index = 0;

                recvword = pullUsbRxBuf();
                if (recvword)  data1 = (uint8_t)(recvword & 0x00FF);
                else index = 0;

                recvword = pullUsbRxBuf();
                if (recvword)  data2 = (uint8_t)(recvword & 0x00FF);
                else index = 0;

                if (index) MIDI_Interpreter(index, status, data1, data2);
            }
        }
    }
}


//##########################################################################
// Misc Low Level Routines
//************************

void msDelay(uint16_t count)   // *******************************************
//blocking delay by 'count' ms
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT) IWDG->KR = 0x0000AAAA;
                                    //feed watchdog, timeout ca. 1-2 seconds;
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in main loop and pullUsbRxBuf()
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
uint8_t classChange;
static uint8_t secCount;

//system relevant, but not extremely time critical services
//are handled quasi in the background

//high speed part:
    receiveUDP();   //Art-Net/UDP is the only Ethernet feature

#ifndef DMX_DEBUG
//USART2 is DMX transmitter, timing handled by polling
    if (USART2->ISR & USART_ISR_TXE) {
        USART2->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
        gdDmxTxIndex++;
        if (gdDmxTxIndex >= (gdDmxMax + 1))  DmxHeader() ; //>= new 30July20
    }
#endif

//lower speed services:  **********************************

    if (gSYSCOUNT == 0)  {
        gSYSCOUNT = 20;         //restart 50Hz service interval

        IWDG->KR = 0x0000AAAA;  //fed watchdog - timeout ca. 1-2 seconds

//handle DMX faders
        if (gdFaderBusy)  fadeProcess();

//LED control
        if (gRED_LEDCOUNT==0)  GPIOA->BSRR =  GPIO_BSRR_BS_0;
                                                    //PA0 (red LED) ON again
        if (gGREEN_LEDCOUNT==0)  GPIOB->BSRR =  GPIO_BSRR_BS_8;
                                                    //PB8 (green LED) ON again

#ifndef DMX_DEBUG
//very low speed service (1 per sec)    *******************
        secCount ++;
        if (secCount >= 50)  {
            secCount = 0;

//check panel switch or jumper and force new USB enumeration if necessary
//switch open: = (default) Artnet and parallel ASCII commands, Mini DMX via USB
//switch closed: = Artnet and parallel MIDI commands via USB
            classChange = 0;
            if (GPIOA->IDR & GPIO_IDR_14)  {        //then switch is open
                if (guDeviceClass == MIDI_CLASS) {
                    stopUSB();
                    guDeviceClass = CDC_CLASS;
                    classChange = 0xFF;
                }
            }
            else if (guDeviceClass == CDC_CLASS) {
                stopUSB();
                guDeviceClass = MIDI_CLASS;
                classChange = 0xFF;
            }
            if (classChange)  {
                GPIOA->BSRR =  GPIO_BSRR_BR_0;          //PA0 (red LED) OFF
                GPIOB->BSRR =  GPIO_BSRR_BR_8;          //PB8 (green LED) OFF
                startUsb();
                GPIOA->BSRR =  GPIO_BSRR_BS_0;          //PA0 (red LED) ON again
                GPIOB->BSRR =  GPIO_BSRR_BS_8;          //PB8 (green LED) ON
            }
        }
#endif
    }
}

void startUsb(void)   //*****************************************************
// **************************************************************************
{
    GPIOA->BSRR = GPIO_BSRR_BR_3;       //disable DMX driver
    initUSB();

    gTIMEOUT = 5000;                    //5 seconds
    while (gTIMEOUT)  {
        IWDG->KR = 0x0000AAAA;          //feed watchdog
        if (guConfigured)  break;
    }
    GPIOA->BSRR = GPIO_BSRR_BS_3;       //enable DMX driver

    if (guDeviceClass == MIDI_CLASS) geArtnetOn = 0; //may be unblocked manually
    else  geArtnetOn = 1;               //default 1= don't block Artnet, 0=block
}


void popFlashPage17(void)   //***********************************************
// **************************************************************************
{
uint16_t  *flashPtr, *comParamPtr, k;

    flashPtr = (uint16_t *)PAGE17;
    comParamPtr = (uint16_t *) &(gComParam.VidPid);
    for (k=0;k<20;k++)  {
        *comParamPtr = ~(*flashPtr++);
        comParamPtr++;
    }
//if flash item is empty, initialize default
 //   if (gComParam.VidPid == 0) gComParam.VidPid = 0x00000000;
    if (gComParam.MacHi == 0) gComParam.MacHi = MACHI_DEFAULT;
    if (gComParam.MacLo == 0) gComParam.MacLo = MACLO_DEFAULT;
    if (gComParam.Gateway == 0) gComParam.Gateway = GATEWAY_DEFAULT;
    if (gComParam.LocalIP == 0) gComParam.LocalIP = LOCALIP_DEFAULT;
    if (gComParam.Subnet == 0) gComParam.Subnet = SUBNET_DEFAULT;
    if (gComParam.VidPid == 0) {
        gVidShadow = VID_DEFAULT;
        gPidShadow = PID_DEFAULT;
    }
    else  {
        gVidShadow = (uint16_t)(gComParam.VidPid >> 16);
        gPidShadow = (uint16_t)(gComParam.VidPid & 0x0000FFFF);
    }

}


void pushFlashPage17(void)   //**********************************************
// **************************************************************************
{
uint16_t  *flashPtr, *comParamPtr, k;

//unlock flash
#define KEY1    0x45670123              //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR & FLASH_SR_BSY) != 0);  //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page25
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = PAGE17;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    else {
        sendUsbX32(FLASH->SR);                      //better error management ?
        sendUsbB('?');
    }
    FLASH->CR &= ~FLASH_CR_PER;

    gComParam.VidPid =  ((uint32_t)gVidShadow << 16) | (uint32_t)gPidShadow;

//write mirror into page25.This processor allows only half-word flash write
    flashPtr = (uint16_t *)PAGE17;
    comParamPtr = (uint16_t *) &(gComParam.VidPid);
    FLASH->CR |= FLASH_CR_PG;
    for (k=0;k<20;k++)  {
        *flashPtr = ~(*comParamPtr++);
        flashPtr++;
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        else {
            sendUsbX32(FLASH->SR);                  //better error management ?
            sendUsbB('!');
        }
    }
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}


void popPresetPage(uint8_t presetNum)   //************************************
// reloads a 512 byte backup of DMXBUF stored (assumed) as one's complements
// so a virgin flash initializes all DMX channels with zero level
// **************************************************************************
{
uint8_t  k;
uint32_t  *flashPtr, *dmxPtr;

    flashPtr = (uint32_t *)(PAGE18 + presetNum*0x400);
    dmxPtr = (uint32_t *) &(gdFINALMEM[0]);
    for (k=0;k<130;k++)  {
        *dmxPtr = ~(*flashPtr++);
        dmxPtr++;
    }
    if (presetNum == 0) {   //fadetime and DMX cycle only updated by preset no.0
        gdDmxMax =
        (uint16_t)(gdFINALMEM[DMX_BUFLEN+2] << 8) | gdFINALMEM[DMX_BUFLEN+3];
        if (gdDmxMax == 0) gdDmxMax = 512;	//bugfix 27July20
		if (gdDmxMax <24 ) gdDmxMax = 24;   //shortest allowed DMX cycle
        gdFadeTime =
        (uint16_t)(gdFINALMEM[DMX_BUFLEN+4] << 8) | gdFINALMEM[DMX_BUFLEN+5];
        if (gdFadeTime > 127) gdFadeTime = 127;
    }

    fade2Preset();
}


void pushPresetPage(uint8_t presetNum)   //**********************************
// saves the content pf DMXBUF permanently as one's complements
// **************************************************************************
{
uint16_t  *flashPtr, *dmxPtr, k;

//unlock flash
#define KEY1    0x45670123              //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR & FLASH_SR_BSY) != 0) ; //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase preset page
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = (PAGE18 + presetNum*0x400) ;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    else {
        sendUsbX32(FLASH->SR);                      //better error management ?
        sendUsbB('!');
    }
    FLASH->CR &= ~FLASH_CR_PER;

//Fadetime and DMX cycle length are stored together with DMX data
    gdDMXBUF[DMX_BUFLEN+2] = (gdDmxMax>>8);         //insert gdDmxMax
    gdDMXBUF[DMX_BUFLEN+3] = (gdDmxMax& 0x00FF);
    gdDMXBUF[DMX_BUFLEN+4] = (gdFadeTime>>8);       //insert gdFadeTime
    gdDMXBUF[DMX_BUFLEN+5] = (gdFadeTime& 0x00FF);

//write DMX buffer into Page (PAGE18 + presetNum*pagesize)
    flashPtr = (uint16_t *)(PAGE18 + presetNum*0x400);
    dmxPtr = (uint16_t *) &(gdDMXBUF[0]);
    FLASH->CR |= FLASH_CR_PG;
    for (k=0;k<260;k++)  {
        *flashPtr = ~(*dmxPtr++);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        else {
            sendUsbX32(FLASH->SR);                  //better error management ?
            sendUsbB('?');
        }
        flashPtr++;
    }
    FLASH->CR &= ~FLASH_CR_PG;

//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}

#ifdef DMX_DEBUG  //else USART2 is opened in dmx.c, startDmxTx()
void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
    RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ;   // USART2 clock enable

	GPIOA->MODER &= ~(GPIO_MODER_MODER2);
	GPIOA->MODER |= (GPIO_MODER_MODER2_1); //Set pin2 in alternate function mode
                                            // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

    GPIOA->BSRR = GPIO_BSRR_BS_3;           //set bit PA3 (enable DMX driver)
    GPIOA->MODER &= ~GPIO_MODER_MODER1;     //PortA1 input in this special case

//USART2 is AF1
    GPIOA->AFR[0] &= 0xFFFFF0FF;
    GPIOA->AFR[0] |= 0x00000100;

    USART2->BRR =0x00D0;        //115200Baud: 13.0 at 24 MHz APB1 Clock
    USART2->CR2 = 0;			//default:1 Stopbit
    USART2->CR3 = 0;			//default:kein Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_TE;      //only TX, 8Bit, no Parity
}


void sendB2(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}


void sendX8(uint8_t data)  // ********************************************
// **************************************************************************
{
uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendB2(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendB2(temp);
    sendB2(0x20);
}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************

{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendB2(asciiNibble);
}


void sendX32(uint32_t data) // **********************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendUsbB(0x20);
}
#endif


void sendUsbString(const uint8_t * str) // ***********************************
// ***************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendUsbB(byte);
}


void sendUsbX8(uint8_t data)  // ********************************************
// **************************************************************************
{
uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbB(temp);
    sendUsbB(0x20);
}

void sendUsbN8(uint8_t data) // *********************************************
// send 1 byte in hex format (without leading 0x), WITHOUT TRAILING SPACE
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbB(temp);
}


void sendUsbNibble(uint8_t data) // ********************************************
// **************************************************************************

{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendUsbB(asciiNibble);
}


void sendUsbX16(uint16_t data) // *******************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sendUsbNibble(nibble);
    sendUsbB(0x20);
}


void sendUsbX32(uint32_t data) // **********************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendUsbNibble(nibble);
    sendUsbB(0x20);
}


void sendUsbN32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendUsbB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendUsbB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendUsbB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendUsbB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendUsbB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendUsbB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendUsbB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendUsbB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendUsbB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendUsbB((uint8_t)quotient + 0x30);
ONES:
    sendUsbB((uint8_t)data + 0x30) ;
}


void sendUsbS32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendUsbN32(data);
    sendUsbB(0x20);
}


void sendUsbS16(uint16_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//up to  +/- 32767
// **************************************************************************
{
uint16_t quotient;

    if (data >= 0x8000) {
        sendUsbB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    quotient = data/10000;
    data = data - quotient*10000;
    sendUsbB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendUsbB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendUsbB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendUsbB((uint8_t)quotient + 0x30);
ONES:
    sendUsbB((uint8_t)data + 0x30) ;
}


// ##########################################################################

// Interrupt Handlers
//*******************

void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
    if ((gSYSCOUNT > 0)&&(gSYSCOUNT < 0x8000)) gSYSCOUNT--; //controls servSystem()
    else gSYSCOUNT = 0;             //double safetiy against underflow

    if (gRED_LEDCOUNT > 0) gRED_LEDCOUNT--;
    if (gGREEN_LEDCOUNT > 0) gGREEN_LEDCOUNT--;
    if (gDELAYCOUNT > 0) gDELAYCOUNT--;
    if (gTIMEOUT > 0) gTIMEOUT--;   //to be used in any subroutine
}


