// USB specific routines for USB+Ethernet/DMX Generator  *******************
//
// Simple DMX512(incl. Art-Net) controller for STM32F042F6 (20pin TSSOP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.c", "main.h", "dmx.c", "ether.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics and EMBITZ.org
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "usb_regs.h"
#include "globaldefines.h"

#define EP0_MAXLEN   0x40

const uint8_t CDC_DEVICEDESCRIPTOR [0x12] =
{
0x12,                   //length of this descriptor=18 bytes
1,                      //descriptor type 	DEVICE Descriptor = 1
0x00,0x02,              //bdcUSB current rev of USB spec= 0x0200 = USB2.00
0,                      //bDeviceClass		Device defined at Interface level
0,       		        //bDeviceSubClass 	= 0x00 unused
0,                      //bDevice Protocol 	= 0x00, unused
EP0_MAXLEN,             //bMaxPacketSize0 for endpoint 0
//if your device shall appear as "STMicroelectronics Virtual Com Port"
VID_DEFAULT & 0x00FF,VID_DEFAULT>>8,
                        //idVendor 		VID= 0x0483
PID_DEFAULT & 0x00FF,PID_DEFAULT>>8,
                        //idProduct 	PID= 0x5740
REVNUM,HARDWARE_VERSION, //bcd_Device	release code
0,                      //iManufacturer index of manufacturer string descriptor (unused)
0,    		            //iProduct	    index of product name string descriptor (unused)
0,                      //iSerialNumber	serial number (unused)
1     	                //bNumConfigurations = 1 only one available
};


const uint8_t CDC_CONFIGDESCRIPTOR [67] =
{
9,                      //bLength		    length of this descriptor = 9 bytes
2,          	        //bDescriptor type	CONFIGURATION descriptor = 2
67,0,        			//wTotalLength	    total length of configuration block=67bytes
2,                      //bNumInterfaces    two interfaces incl this descriptor
1,      		        //bConfigurationValue ID of this configuration = 1
0,                      //iConfiguration	index of string descriptor:unused=0
0x80,                   //bmAttributes		BUS-powered - no remote wakeup
99,         	        //MaxPower			entry*2 = 198mA

// CDC_INTERFACE_DESCRIPTOR
9,                      //bLength       	length of this descriptor = 9 bytes
4,                      //bDescriptorType	INTERFACE descriptor type=4
0,          	        //bInterfaceNumber	index of this interface =0
0,                      //bAlternateSetting	index of this setting =0
1,		                //bNumEndpoints 	(Notification endpoint) = 1
2,                      //bInterfaceClass 	= 2 (COMMUNICATION)
2,       		        //bInterfaceSubClass = 2 (ABSTRACT_CONTROL_MODEL)
0,                      //bInterfaceProtocol = 0 (unused)
0,        	            //iInterface 		= 0 (unused)

//COMMUNICATION DEVICE specific functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,       	        //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
0,                      //bDescriptortSubtype HEADER subtype
0x10,1,                 //bcdADC	=0x0100 revision of class specific descriptors

//CALL MANAGEMENT functional Descriptor
5,                      //bLength           length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
1, 			            //bDescriptorSubtype
3,                      //bmCapabilities	Device handles call management by itself
                        //ia DataInterface
						//i.e. in this case: no call management required
0,                     	//bDataInterface:	Data interface no.0 (see below) associated
//1,                     	//bDataInterface:	Data interface no.0 (see below) associated

//ABSTRACT CONTROL management functional Descriptor
4,                      //bLength			length of this descriptor = 4 bytes
0x24,         	        //bDesriptorType	CS_INTERFACEdescriptor type = 0x24
2,                      //bDescriptorSubtype
7,				        //bmCapabilities    Send_Break,Set_Line_Coding... etc

//UNION functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACE descriptor
6,                      //bDescriptorSubtype UNION FUNCTIONAL descriptor
0,		       	        //bMasterInterface
1,                      //bSlaveInterface0

//Notification standard endpoint descriptor
7,  			        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x83,			        //bEndpointAddress	no.3, IN
3,                      //bmAttributes 	INTERRUPT = 3
16,0,                    //wMaxPacketSize	=16
64,			            //interval in ms    =64

//DATA CLASS INTERFACE descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,				        //bDescriptorType	INTERFACE descriptor type
1,                      //bInterfaceNumber
0,				        //bAlternateSetting
2,                      //bNumEndpoints
0x0A,       			//bInterfaceClass	Data Class
0,                      //bInterfaceSubclass unused
0,				        //bInterfaceProtocol no specific protocol
0,                      //iInterface 		unused

//DATA BULK IN endpoint standard descriptor
7,				        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x81,			        //bEndpointAddress	nr.1, IN
2,                      //bmAttributes		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
0,				        //bInterval			=0:ignore for bulk transfer (vorher:1ms)

//DATA BULK OUT endpoint standard descriptor
7,                      //bLength			length of this descriptor = 7 bytes
5,				        //bDescriptorType	ENDPOINT
2,                      //bEndpointAddress	nr.=2, OUT
2,				        //bmAttributes 		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
0				        //bInterval			=0:ignore for bulk transfer (vorher:1ms)
};

//Product String Descriptor index 1
const uint8_t PRODUCT_STRINGDESCRIPTOR [38] =
{
38,                     //bLength = 30+2  Descriptor length in Bytes
0x03,                   //bDescriptorType = 0x03, Constant String = 3
'U',0,'S',0,'B',0,'/',0,'D',0,'m',0,'x',0,'+',0,'A',0,'r',0,'t',0,'n',0,'e',0,'t',0,' ',0,'G',0,'e',0,'n',0
                        //bString (Unicode 16Bit)
};

const uint8_t LINE_CODING_115200 [7] =
{
0x00,0xC2,0x01,0x00,0,0,8  //0x0001C200=decimal 115200, rest describes 1N8
//the baud rate is faked, the microcontroller accepts any and sets its own for the USART.
//but 115200 seems to be the best default at host side.
//Else enter your baudrate 32bit formatted.
};


const uint8_t MIDI_DEVICEDESCRIPTOR[0x12] =
{
18,        	            //bLength			length of this descriptor=18 bytes
1,                     	//descriptor type 	DEVICE Descriptor = 1
0x00,0x02,              //bdcUSB current rev of USB spec= 0x0200 = USB2.00
0,            		    //bDeviceClass		Device defined at Interface level
0,                     	//bDeviceSubClass 	= 0x00 unused
0,	    	            //bDevice Protocol 	= 0x00, unused
EP0_MAXLEN,	            //bMaxPacketSize0	Endp.0 = 64
VID_DEFAULT & 0x00FF,VID_DEFAULT>>8,
                        //idVendor 		    VID= 0x0483
PID_DEFAULT & 0x00FF,(PID_DEFAULT>>8) + 1,
                        //idProduct PID=0x5741 ->Pid(CDC)+1 !else host problem!
REVNUM,HARDWARE_VERSION, //bcd_Device		release code
0,   		            //Manufacturer		index of manufacturer string descriptor (unused)
1,                   	//iProduct			Index of product name string descriptor (=1)
0,     		            //iSerialNumber		serial number (unused)
1                    	//bNumConfigurations = 1 only one available
};


const uint8_t MIDI_CONFIGDESCRIPTOR[0x65] =  //dez101
{
9,                      //bLength			length of this descriptor = 9 bytes
2,                      //bDescriptorType	CONFIGURATION descriptor type = 2
0x65,0,			        //wTotalLength		length of configuration block =101
                        //					incl this descriptor
2,      		        //bNumInterfaces    two interfaces
1,						//bConfigurationValue ID of this configuration
0,                      //iConfiguration	index of string descriptor:unused=0
0x80,                    //bmAttributes		BUS-powered - no remote wakeup
99,                      //MaxPower 		entry*2 = 198mA
//Standard Audio Class Interface Descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,          	        //bDescriptorType	INTERFACE descriptor type=4
0,                      //bInterfaceNumber	index of this interface
0,	                    //bAlternativeSetting index of this setting
0,	                    //bNumEndpoints		= 0 (EP0 only)
1,       		        //bInterfaceClass	= 0x01 (AUDIO)
1,                      //bInterfaceSubClass = 0x01 (AUDIO_CONTROL)
0,         	            //bInterfaceProtocol = 0 (unused)
0,                      //iInterface 		= 0 (unused)
//Class-specific AudioControl Interface descriptor
9,       	            //bLength			length of this descriptor = 9 bytes
0x24,                   //bDescriptorTye	CS_INTERFACEdescriptor type = 0x24
1,       		        //bDescriptorSubtyte HEADER subtype
0,1,                    //bcdADC			=0x0100 revision of class specific descriptors
9,0,                    //wTotalLength		= 9
1, 			            //bInCollection		number of streaming interfaces
1,   			        //baInterfaceNr(1)	MIDIStreaming interface no.1
						//					belongs to this Audio Control
//Standard MIDI Streaming Interface Descriptor
9,                      //bLength	        length of this descriptor = 9 bytes
4,         	            //bDescriptorType	INTERFACE descriptor type = 0x04
1,                      //ibInterfaceNumber	index of this interface
0,				        //bAlternativeSetting index of this setting
2,						//bNumENdpoints		2 endpoints
1, 			            //bInterfaceClass	AUDIO
3, 						//bInterfaceSubclass MIDISTREAMING
0,          	        //bInterfaceProtocol = 0 (unused)
0,                       //iInterface 		= 0 (unused)
//Class specific MIDI Streaming Interface Descriptor
7,                      //bLength 			length of this descriptor = 7 bytes
0x24,                   //bDescriptorType	CD_INTERFACE descriptor
1,       	            //bDescriptorSubtye	MS_HEADER subtype
0,0x01,                 //bcdADC			Revision of this class specification 0x0100
0x41,0x00,		        //wTotalLength		total size of class-specific descriptors
//MIDI IN Jack descriptor (embedded)
6,                      //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
1,				        //bJackType			EMBEDDED
1,						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI IN Jack descriptor (external)
6,			            //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
2,				        //bJackType			EXTERNAL
2, 						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI OUT Jack descriptor (embedded)
9,			            //bLength           length of this descriptor = 9 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
3, 						//bDescriptorSubtype MIDI_OUT_JACK subtype
1,				        //bJackType			EMBEDDED
3, 						//bJackID			ID of this jack
1,				        //bNrInputPins		= 1
2,						//BaSourceID(1)		this jack is connected to jack with ID 2
1,				        //BaSourcePin(1)	there connected to pin 1
0,   					//iJack 			unused
//MIDI OUT Jack descriptor (external)
9,			            //length of this descriptor = 9 bytes
0x24,					//bDescriptorType	CS_INTERFACE descriptor
3,				        //bDescriptorSubtype MIDI_OUT_JACK subtype
2, 						//bJackType			EXTERNAL
4,				        //bJackID			ID of this jack
1,						//bNrInputPins		= 1
1,				        //BaSourceID(1)		this jack is connected to jack with ID 1
1,						//BaSourcePin(1)	there connected to pin 1
0,  			        //iJack 			unused
//Standard BULK IN endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,			            //bDescriptorType	ENDPOINT descriptor
0x81,					//bEndpointAddress	IN endpoint no.1
2,			            //bmAttributes		BULK DATA, not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK IN endpoint descriptor
5,			            //bLength			length of this descriptor = 5 bytes
0x25,					//bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI OUT jacks
3,				        //BaAssocJackID(1)	associated with embedded MIDI OUT no.3
//Standard BULK OUT endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,				        //bDescriptorType	ENDPOINT descriptor
2,   					//bEndpointAddress	OUT endpoint no.2
2,			            //bmAttributes		BULK DATA,not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK OUT endpoint descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x25,			        //bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI IN jacks
1				        //baAssocJackID(1)	associated with embedded MIDI IN no.1
};


//absolute addresses of USB_EPnR registers
#define USB_EP0R    ((volatile uint16_t *)(RegBase))    //adopted from  usb_regs.h
#define USB_EP1R    ((volatile uint16_t *)(RegBase+4))
#define USB_EP2R    ((volatile uint16_t *)(RegBase+8))
#define USB_EP3R    ((volatile uint16_t *)(RegBase+0x0C))
#define USB_EP4R    ((volatile uint16_t *)(RegBase+0x10))
#define USB_EP5R    ((volatile uint16_t *)(RegBase+0x14))
#define USB_EP6R    ((volatile uint16_t *)(RegBase+0x18))
#define USB_EP7R    ((volatile uint16_t *)(RegBase+0x1C))

#define CLEAR_CTR_RX_MASK  (EPREG_MASK & 0x7FFF)  //=0x0F8F
#define CLEAR_CTR_TX_MASK  (EPREG_MASK & 0xFF7F)  //=0x8F0F
#define EP0_TX_OFFSET   0x40
#define EP0_RX_OFFSET   0x80
#define EP1_TX_OFFSET  0x100
#define EP2_RX_OFFSET  0x200
#define EP3_TX_OFFSET  0x300


// ***************************************
extern uint16_t gSYSCOUNT, gRED_LEDCOUNT;

extern struct PARAM{
uint32_t VidPid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t Gateway;
uint32_t ArtnetIP;
uint32_t ArtSubnet;
uint32_t ArtUniverse;
uint32_t Channel;
} gComParam;  //10 words

extern uint16_t gVidShadow, gPidShadow;

extern void msDelay(uint16_t);
extern void servSystem(void);
#ifdef DMX_DEBUG
extern void sendB2(uint8_t);
extern void sendX8(uint8_t);
#endif

// ***************************************

//Endpoint 0
uint16_t  *gUSB_ADDR0_TX, *gUSB_COUNT0_TX, *gUSB_ADDR0_RX, *gUSB_COUNT0_RX ;
//Endpoint 1 is single buffer IN
uint16_t  *gUSB_ADDR1_TX, *gUSB_COUNT1_TX;
//Endpoint 2 is single buffer OUT
uint16_t  *gUSB_ADDR2_RX, *gUSB_COUNT2_RX;
//Endpoint 3 is single buffer IN
uint16_t  *gUSB_ADDR3_TX, *gUSB_COUNT3_TX;

#define USB_BUFLEN  1120            //division by 8 must be full integer
uint8_t gUSB_TXBUF[USB_BUFLEN/2];   //for EP1 IN
volatile uint8_t *gUSB_TXHEAD=0,*gUSB_TXTAIL=0,*gUSB_TXBUFTOP=&(gUSB_TXBUF[USB_BUFLEN/2]);
uint8_t gUSB_RXBUF[USB_BUFLEN];     //for EP2 OUT, CDC_CLASS
volatile uint8_t *gUSB_RXHEAD=0,*gUSB_RXTAIL=0,*gUSB_RXBUFTOP=&(gUSB_RXBUF[USB_BUFLEN]);

uint16_t * guBTABLE_ENTRY;
volatile uint16_t guAddress;

//USB Std Request parameters
volatile uint8_t   guRequestType;
volatile uint8_t   guRequest;
volatile uint16_t  guValue;
volatile uint8_t   guDescriptor;
volatile uint16_t  guIndex;
volatile uint16_t  guLength;
// USB packet handling
volatile uint8_t   guSetupAction;   //action to be performed in servEP0_In() routine
volatile uint8_t   guTx0RemainCnt;  //remaining number of bytes to be transmitted at EP0
const uint8_t      *guInSendPtr;    //pointer to storage of next EP0 data to be transmitted
uint8_t            guDeviceClass, guConfigured, guLockEP2;
uint8_t            *guEP2Ptr;       //pointer to EP2_OUT fifo

void initUSB(void);
void subInitUSB(void);

void USB_IRQHandler(void);
void servEP0_Out(void);
void servEP0_In(void);
void prepareDescriptor(void);
void sendZLP_EP0(void);
void stall_EP0(void);
void servEP1_In(void);
void sendUsbB(uint8_t);
void servEP2_Out(void);
uint16_t pullUsbRxBuf(void);
void servEP3_In(void);
uint8_t packMidiMsg(uint8_t len);
uint32_t pullMidiMsg(void);

//##########################################################################

void stopUSB(void)   // *****************************************************
//  ***************************************************************************
{
    guConfigured = 0;

    GPIOA->MODER&=~(GPIO_MODER_MODER11|GPIO_MODER_MODER12);
    NVIC_DisableIRQ(USB_IRQn);
    RCC->APB1RSTR |= RCC_APB1RSTR_USBRST;       //reset USB
    RCC->APB1ENR &= ~RCC_APB1ENR_USBEN;         //disable USB clock
    msDelay(100);
}


void initUSB(void)  // ****************************************************
// *************************************************************************
{
uint32_t k;
uint16_t * tempPtr;

    RCC->APB1RSTR |= RCC_APB1RSTR_USBRST;
    msDelay(2);
    RCC->APB1RSTR &= ~RCC_APB1RSTR_USBRST;
    RCC->CFGR3 |= RCC_CFGR3_USBSW;		    //select PLL as USB clock
 //   RCC->CFGR3 &=  ~RCC_CFGR3_USBSW;		    //select HSI48 as USB clock
    RCC->APB1ENR |= RCC_APB1ENR_USBEN;

//as a precaution, init PA11,12 as high impedance inputs
//GPIOA power enabled in initUSART2
    GPIOA->MODER &= ~(GPIO_MODER_MODER11_0 |    // config pin 11,12 as input
                      GPIO_MODER_MODER11_1 |
                      GPIO_MODER_MODER12_1 |
                      GPIO_MODER_MODER12_0 ) ;

    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2MHz by default
    GPIOA->PUPDR &=  ~(GPIO_PUPDR_PUPDR11_0 |   //config pin 11,12  HiZ
                       GPIO_PUPDR_PUPDR11_1 |
                       GPIO_PUPDR_PUPDR12_0 |
                       GPIO_PUPDR_PUPDR11_1 ) ;

    tempPtr= (uint16_t *) RegBase;
    for (k=0; k<0x10; k++)  *(tempPtr+k) = 0;
    for (k=0x22; k<0x30; k++)  *(tempPtr+k) = 0;

// init endpoint registers
    *BCDR |= BCDR_DPPU;      //enable internal pull-up at USB-DP
    *CNTR = CNTR_FRES;       //switch USB power on but keep forced reset state
    msDelay(200);
    *CNTR = 0x0000;          //terminate forced reset
    *ISTR = 0x0000;          //remove spurious pending interrupts

    subInitUSB();

    NVIC_EnableIRQ(USB_IRQn);
}

void subInitUSB(void)  // ****************************************************
// *************************************************************************
{

    *DADDR = DADDR_EF;  // =0x80 and addr=0
    guAddress = 0;
    *BCDR |= BCDR_DPPU; //enable internal pull-up at USB-DP
    *BTABLE = 0x00;     //init USB_BTABLE register
                        //(address relative to start of dedicated packet memory)

//init "buffer description table"
    guBTABLE_ENTRY = (uint16_t *) (PMAAddr + *BTABLE);

//because serial speed is much less than USB speed,only single buffering is used
    gUSB_ADDR0_TX = guBTABLE_ENTRY;
    gUSB_COUNT0_TX = guBTABLE_ENTRY+1;
    gUSB_ADDR0_RX = guBTABLE_ENTRY+2;
    gUSB_COUNT0_RX = guBTABLE_ENTRY+3;
    gUSB_ADDR1_TX = guBTABLE_ENTRY+4;
    gUSB_COUNT1_TX = guBTABLE_ENTRY+5;
    gUSB_ADDR2_RX = guBTABLE_ENTRY+10;
    gUSB_COUNT2_RX = guBTABLE_ENTRY+11;
    if (guDeviceClass == CDC_CLASS)  {
        gUSB_ADDR3_TX = guBTABLE_ENTRY+12;
        gUSB_COUNT3_TX = guBTABLE_ENTRY+13;
    }
//entries for double buffers initalized formally:
    *gUSB_ADDR0_TX = EP0_TX_OFFSET; //corresponds with ref.manual p.871:
                                    //ADDR0_TX, buffer size 64 bytes
    *gUSB_COUNT0_TX = 0;            //COUNT0_TX
    *gUSB_ADDR0_RX = EP0_RX_OFFSET; //ADDR0_RX, buffer size 64 bytes
    *gUSB_COUNT0_RX = 0x8400;       //COUNT0_RX - binary 1 000 0100  0000 0000
    *gUSB_ADDR1_TX = EP1_TX_OFFSET; //ADDR1_TX_0, buffer size 128 bytes, IN endpoint
    *gUSB_COUNT1_TX = 0;             //COUNT1_TX_0
    *(guBTABLE_ENTRY+6) = EP1_TX_OFFSET; //ADDR1_TX_1, buffer size 128 bytes
    *(guBTABLE_ENTRY+7) = 0;              //COUNT1_TX_1
    *(guBTABLE_ENTRY+8) = EP2_RX_OFFSET; //ADDR2_2RX_1, buffer size 128 bytes
    *(guBTABLE_ENTRY+9) = 0;         //COUNT2_RX_
    *gUSB_ADDR2_RX = EP2_RX_OFFSET; //ADDR2_RX_0, buffer size 64 bytes, OUT endpoint
    *gUSB_COUNT2_RX = 0x8400;        //COUNT2_RX_0
    if (guDeviceClass == CDC_CLASS)  {
        *gUSB_ADDR3_TX = EP3_TX_OFFSET; //ADDR3_TX_0, buffer size 32 bytes, IN endpoint
        *gUSB_COUNT3_TX = 0;             //COUNT3_TX_0
        *(guBTABLE_ENTRY+14) = EP3_TX_OFFSET; //ADDR3_TX_1, buffer size 32 bytes
        *(guBTABLE_ENTRY+15) = 0;        //COUNT3_TX_1
    }
// init endpoint registers
    *USB_EP0R =  0x3220;  // Control, everything prepared for first Setup token
    *USB_EP1R =  0x3031;        // Bulk IN
    *USB_EP2R =  0x3032;        // Bulk OUT
    if (guDeviceClass == CDC_CLASS)  *USB_EP3R =  0x3633;    // Interrupt IN

    *CNTR = CNTR_CTRM | CNTR_ERRM | CNTR_WKUPM | CNTR_SUSPM | CNTR_RESETM \
            | CNTR_SOFM  | CNTR_ESOFM;  //set appropriate interrupt mask

    guTx0RemainCnt = 0;

    guConfigured = 0;
    gUSB_TXHEAD = gUSB_TXBUF;       //flush EP1 application level buffer
    gUSB_TXTAIL = gUSB_TXBUF;
    gUSB_RXHEAD = gUSB_RXBUF;       //flush EP2 application level buffer (CDC)
    gUSB_RXTAIL = gUSB_RXBUF;
}

//##########################################################################

//the rest is done by Interrupt Handling
//***************************************

void USB_IRQHandler(void)   // ***********************************************
// ***************************************************************************
{

volatile uint16_t  EPindex, ioDir, wIstr;  //EPstatus,

    wIstr = *ISTR;
    if ((*ISTR & ISTR_CTR) != 0)           //if "Correct Transfer"
    {
        EPindex = wIstr & 0x000F;
        ioDir = wIstr & 0x0010;

//check project specific endpoints

        if (EPindex == 1)  servEP1_In();
// **********************

        else if (EPindex == 2)  servEP2_Out();
// ***************************

        else if (EPindex == 0 ) {
// *****************************
 //check if SETUP and/or CTR_RX are set
            if (ioDir == 0)  servEP0_In();   //ioDir==0, i.e. IN packet
            else servEP0_Out();
        }

        else if ((EPindex == 3) && (guDeviceClass == CDC_CLASS))  servEP3_In();
// ****************************

        else {;}        //ignore
// *****************************

    }
    else if (*ISTR & ISTR_RESET)  {
        subInitUSB();
        *ISTR &= ~ISTR_RESET;
    }
    else if (*ISTR & ISTR_SUSP)  *ISTR &= ~ISTR_SUSP;   //clear SUSP flag
    else *ISTR = 0x807F;                     //reset all 'rc_w0' type flags
}


void servEP0_Out(void)   // **************************************************
// ***************************************************************************
{
uint16_t len, k;
volatile uint16_t regVal;
uint8_t TEMP_CONFIG_BUF[0x12];
uint8_t * tempPtr;

//check if SETUP is set
    if (*USB_EP0R & EP_SETUP)  {
//SETUP packet
        *USB_EP0R = *USB_EP0R & CLEAR_CTR_RX_MASK;      //clear CTR_RX
        guRequestType = *((uint8_t*) (PMAAddr + EP0_RX_OFFSET));
        guRequest = *((uint8_t*) (PMAAddr + EP0_RX_OFFSET +1));
        guValue = *((uint16_t*) (PMAAddr + EP0_RX_OFFSET +2));
        guDescriptor = (uint8_t) ((guValue >> 8) & 0x00FF);
        guIndex = *((uint16_t*) (PMAAddr + EP0_RX_OFFSET +4));
        guLength = *((uint16_t*) (PMAAddr + EP0_RX_OFFSET +6));
        guSetupAction = 0;
        *gUSB_COUNT0_RX = 0x8400;                       //neutralize

        if ((guRequestType == 0x21) && (guDeviceClass == CDC_CLASS)) {
                            // special case SET_LINE_CODING or SET_LINE_STATE
            if ((guRequest != 0x20) && (guRequest != 0x22))  {
                stall_EP0();
                return;
            }
            else {
                sendZLP_EP0();
                guConfigured = 0xFF;
            }
        }
        else if ((guRequestType == 0xA1) && (guRequest == 0x21) \
                                        && (guDeviceClass == CDC_CLASS))  {
            guSetupAction = 0xA1;               //special case GET_LINE_CODING
            guTx0RemainCnt = 7;
            if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
            guInSendPtr = LINE_CODING_115200;
            prepareDescriptor();
        }
        else if (((guRequestType & 0x61) == 0) || ((guRequestType & 0x61) == 1))  {
            switch(guRequest)  {
            case 0:             //Get_Status
                break;
            case 0x05:          //Set_Address
                guTx0RemainCnt = 0;
                sendZLP_EP0();
                guAddress = guValue & 0x7F;     //keep new address globally
                guSetupAction = 5;
                break;
            case 0x06:           //Get_Descriptor
                guSetupAction = 6;
                if (guDescriptor == 1)  {
                    guTx0RemainCnt = 0x12;       //18
                   if (guDeviceClass == MIDI_CLASS) {
                        guInSendPtr = MIDI_DEVICEDESCRIPTOR;
                    }
                    else  guInSendPtr = CDC_DEVICEDESCRIPTOR;
//if configured individually, assemble user defined Vid/Pid
                    tempPtr = TEMP_CONFIG_BUF;
                    for (k=0;k<0x12;k++)  *tempPtr++ = *guInSendPtr++;

                    if (gComParam.VidPid !=0) { //Vid/Pid in flash is not virgin
                        TEMP_CONFIG_BUF[8] = (uint8_t)((gComParam.VidPid>>16) & 0x000000FF);
                        TEMP_CONFIG_BUF[9] = (uint8_t)((gComParam.VidPid>>24) & 0x000000FF);
                        TEMP_CONFIG_BUF[10] = (uint8_t)((gComParam.VidPid) & 0x000000FF);
                        TEMP_CONFIG_BUF[11] = (uint8_t)((gComParam.VidPid>>8) & 0x000000FF);
                        if (guDeviceClass == MIDI_CLASS) TEMP_CONFIG_BUF[10] += 1;
                                //the MIDI PID must not be equal the CDC PID,
                                //else the Host may get confused
                    }
                    guInSendPtr = TEMP_CONFIG_BUF;
                }
                else if (guDescriptor == 2)  {

                    if (guDeviceClass == MIDI_CLASS) {
                        guTx0RemainCnt = 0x65;       //101
                        guInSendPtr = MIDI_CONFIGDESCRIPTOR;
                    }
                    else  {
                        guTx0RemainCnt = 0x43;       //67
                        guInSendPtr = CDC_CONFIGDESCRIPTOR;
                    }
                    if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
                }
                else if (guDescriptor == 3)  {
                    guTx0RemainCnt = 38;
                    if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
                    guInSendPtr = PRODUCT_STRINGDESCRIPTOR;
                }
                else  {
                    stall_EP0();
                    return;
                }
                prepareDescriptor();
                break;
            case 0x09:          //Set_Configuration
                guTx0RemainCnt = 0;
                sendZLP_EP0();
                break;
            default:
                stall_EP0();      //unsupported: stall
                break;
            }
        }
        else  stall_EP0();
    }

    else   {
//OUT packet, not SETUP
//*********************
        len = *gUSB_COUNT0_RX & 0x03FF;
        if(len==0)   {
//make EP RX "VALID":
            if (*USB_EP0R & 0x8000) *USB_EP0R = *USB_EP0R & CLEAR_CTR_RX_MASK;
            regVal = *USB_EP0R;
            if ((regVal & EPRX_DTOG2)==0) *USB_EP0R = 0xA280; //EPRX_DTOG2=0x2000
            if ((regVal & EPRX_DTOG1)==0) *USB_EP0R = 0x9280; //EPRX_DTOG2=0x1000
//Status stage will be followed by Setup Stage
            if (regVal & EP_DTOG_TX) *USB_EP0R = 0x82C0;      //EP_DTOG_TX=0x0040
        }
    }
}


void prepareDescriptor (void)   // *******************************************
// ***************************************************************************
{
volatile uint16_t regVal;
uint16_t k, temp1, temp2;
uint16_t * EP0InPtr;

//analyse type of request
    EP0InPtr = (uint16_t*) (PMAAddr + EP0_TX_OFFSET);

    if (guTx0RemainCnt > EP0_MAXLEN) {
        *gUSB_COUNT0_TX = EP0_MAXLEN;
        guTx0RemainCnt -= EP0_MAXLEN;
    }
    else   {
       *gUSB_COUNT0_TX =guTx0RemainCnt;
        guTx0RemainCnt = 0;
    }

    for (k=0;k<((*gUSB_COUNT0_TX+1)/2);k++)  {
        temp1 = (uint16_t)*guInSendPtr;
        guInSendPtr++;
        temp2 = (uint16_t)*guInSendPtr;
        guInSendPtr++;
        *EP0InPtr = temp1 | temp2<<8;
        EP0InPtr++;
    }

//make EP0 TX "VALID":
    if (*USB_EP0R & 0x0080)  *USB_EP0R = *USB_EP0R & CLEAR_CTR_TX_MASK; //send ACK
    *gUSB_COUNT0_RX = 0x8400;    //neutralize (see "DCD_EP_PrepareRx()")
    regVal = *USB_EP0R;
    if ((regVal & EPTX_DTOG2)==0) *USB_EP0R = 0x82A0; //EPTX_DTOG2=0x0020
    if ((regVal & EPTX_DTOG1)==0) *USB_EP0R = 0x8290; //EPTX_DTOG2=0x0010
}


void sendZLP_EP0(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal;

    *gUSB_COUNT0_TX = 0;

//make EP TX "VALID":
    if (*USB_EP0R & 0x0080)  *USB_EP0R = *USB_EP0R & CLEAR_CTR_TX_MASK; //send ACK
    *gUSB_COUNT0_TX = 0;
    regVal = *USB_EP0R;
    if ((regVal & EPTX_DTOG2)==0) *USB_EP0R = 0x82A0; //EPTX_DTOG2=0x0020
    if ((regVal & EPTX_DTOG1)==0) *USB_EP0R = 0x8290; //EPTX_DTOG2=0x0010
//Status stage will be followed by Setup Stage
    if (regVal & EP_DTOG_RX) *USB_EP0R = 0xC280;      //EP_DTOG_RX=0x4000
}


void servEP2_Out(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal, fillstate;
uint8_t len;

    if (*USB_EP2R & 0x8000) *USB_EP2R = *USB_EP2R & CLEAR_CTR_RX_MASK;

    guEP2Ptr = (uint8_t *) (PMAAddr + EP2_RX_OFFSET);
    len = *gUSB_COUNT2_RX & 0x03FF;

//restore endpoint buffer into cyclic application level buffer
    while ((len > (guDeviceClass+(guDeviceClass<<1))) && (len < 65 )) { //*3
                                //lower limit:CDC_CLASS:>0, MIDI_CLASS=>3
                                // limits are protection against underflow
        if (guDeviceClass == CDC_CLASS)  {
            *gUSB_RXHEAD++ = *guEP2Ptr ;
            guEP2Ptr++ ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            len-- ;
        }
        else  {
//regular MIDI packets always consist of 4 bytes:
            *gUSB_RXHEAD++ = *guEP2Ptr++ ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            *gUSB_RXHEAD++ = *guEP2Ptr++ ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            *gUSB_RXHEAD++ = *guEP2Ptr++ ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            *gUSB_RXHEAD++ = *guEP2Ptr++ ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            len -= 4;
        }
    }

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD?
    if (gUSB_RXTAIL > gUSB_RXHEAD) {
        fillstate=(uint32_t)gUSB_RXHEAD + USB_BUFLEN -(uint32_t)gUSB_RXTAIL;
    }
    else fillstate = (uint32_t)gUSB_RXHEAD - (uint32_t)gUSB_RXTAIL ;

// lock EP2 interrupt when EP2_BUF contains at least (SIZE-64) bytes
                    // gets unlocked in routine pullUsbRxBuf
                    // as soon as EP2_BUF contains less than (SIZE-64) bytes
                    //this way EP2 always can accept
                    //another packet of 64 bytes without data loss
                    //(similar to CTS/RTS handshake at RS-232)
    guLockEP2 = 0;
    if (fillstate > USB_BUFLEN - 70)  {
        guLockEP2 = 0xFF;
        return;     //i.e. EP RX remains NAKed !! until in gets unlocked
    }

    //make EP RX "VALID" again:
    regVal = *USB_EP2R;
    if ((regVal & EPRX_DTOG2)==0) *USB_EP2R = 0xA082; //EPRX_DTOG2=0x2000
    if ((regVal & EPRX_DTOG1)==0) *USB_EP2R = 0x9082; //EPRX_DTOG2=0x1000
}


uint16_t pullUsbRxBuf(void) // **********************************************
// **************************************************************************
{
volatile uint16_t regVal,recv, fillstate;

    servSystem();   //trigger DMX Header and serve watchdog etc.

    if (guConfigured == 0)  return 0;

// check first if something received. If, then set Bit15. If not, return 0
 //   else if (gUSB_RXTAIL != gUSB_RXHEAD)
    if (gUSB_RXTAIL != gUSB_RXHEAD)
    {
        recv= (uint16_t)((*gUSB_RXTAIL++ & 0x00FF)|0x8000);
        if (gUSB_RXTAIL == gUSB_RXBUFTOP)  gUSB_RXTAIL = gUSB_RXBUF;

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD? then add USBBUFLEN to get the absolute difference
        if (gUSB_RXTAIL > gUSB_RXHEAD) {
            fillstate = (uint32_t)gUSB_RXHEAD + USB_BUFLEN - (uint32_t)gUSB_RXTAIL ;
        }
        else fillstate = (uint32_t)gUSB_RXHEAD - (uint32_t)gUSB_RXTAIL ;

// unlock EP2 interrupt when EP2_BUF contains less than (SIZE-64) bytes
        if ((guLockEP2) && (fillstate < USB_BUFLEN - 70))  {
            guLockEP2 = 0;
//make EP RX "VALID" again:
            regVal = *USB_EP2R;
            if ((regVal & EPRX_DTOG2)==0) *USB_EP2R = 0xA082; //EPRX_DTOG2=0x2000
            if ((regVal & EPRX_DTOG1)==0) *USB_EP2R = 0x9082; //EPRX_DTOG2=0x1000
        }
        GPIOA->BSRR = GPIO_BSRR_BR_0; // red LED OFF
        gRED_LEDCOUNT = 250;  //ms

        return recv;
    }
    else return (uint16_t)0x0000;
}


void servEP0_In(void)   // ***************************************************
// ***************************************************************************
{
volatile uint16_t regVal;

    *USB_EP0R = *USB_EP0R & CLEAR_CTR_TX_MASK;  //sendACK

    if (guSetupAction == 5)  {
        *DADDR = DADDR_EF | guAddress;          // =0x80 plus new addr
        guSetupAction = 0;
        if (guDeviceClass == MIDI_CLASS)  guConfigured = 0xFF;
    }
    if (guTx0RemainCnt == 0) {

//make EP RX "VALID":
        if (*USB_EP0R & 0x8000) *USB_EP0R = *USB_EP0R & CLEAR_CTR_RX_MASK;
                                                          //clear CTR_RX
        regVal = *USB_EP0R;
        if ((regVal & EPRX_DTOG2)==0) *USB_EP0R = 0xA280; //EPRX_DTOG2=0x2000
        if ((regVal & EPRX_DTOG1)==0) *USB_EP0R = 0x9280; //EPRX_DTOG2=0x1000
    }
    else prepareDescriptor();   //prepare next packet of multipacket request
}


void stall_EP0(void)   // ***************************************************
// **************************************************************************
{
volatile uint16_t regVal;

//toggle EP TX "Stalled":
        regVal = *USB_EP0R;
        if ((regVal & EPTX_DTOG1)==0) *USB_EP0R = 0x8290; //EPTX_DTOG2=0x0010
        if ((regVal & EPTX_DTOG2)==1) *USB_EP0R = 0x82A0; //EPTX_DTOG2=0x0020
//toggle EP RX "Stalled":
        if ((regVal & EPRX_DTOG2)==0) *USB_EP0R = 0xA280; //EPRX_DTOG2=0x2000
        if ((regVal & EPRX_DTOG1)==1) *USB_EP0R = 0x9280; //EPRX_DTOG2=0x1000
}


void servEP1_In(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal,temp1,temp2;
uint16_t *EP1Ptr;
uint8_t len;

//clear interrupt flag:
    if (*USB_EP1R & 0x0080)  *USB_EP1R = *USB_EP1R & CLEAR_CTR_TX_MASK; //send ACK

//fill the buffer:
    EP1Ptr = (uint16_t *) (PMAAddr + EP1_TX_OFFSET);
    len = 0;
    while (gUSB_TXTAIL != gUSB_TXHEAD)  {
        temp2 = 0;
        temp1 = (uint16_t)*gUSB_TXTAIL;
        gUSB_TXTAIL++;
        if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
        len++;
        if (len == 64) {
            *EP1Ptr = temp1;
            break;
        }
        if (gUSB_TXTAIL != gUSB_TXHEAD)  {
            temp2 = (uint16_t)*gUSB_TXTAIL;
            gUSB_TXTAIL++;
            if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
            len++;
        }
        *EP1Ptr = temp1 | temp2<<8;
        EP1Ptr++;
        if (len == 64) break;
    }
    *gUSB_COUNT1_TX = len;

//make EP TX "VALID":
    regVal = *USB_EP1R;
    if ((regVal & EPTX_DTOG2)==0) *USB_EP1R = 0x80A1; //EPTX_DTOG2=0x0020
    if ((regVal & EPTX_DTOG1)==0) *USB_EP1R = 0x8091; //EPTX_DTOG2=0x0010
}


void sendUsbB(uint8_t byte)  // ******************************************
// add data byte to EP1 transmit buffer
// **************************************************************************
{
    if (guConfigured == 0)  return;
    *gUSB_TXHEAD++ = byte;
    if (gUSB_TXHEAD == gUSB_TXBUFTOP)  gUSB_TXHEAD=gUSB_TXBUF;
}


void servEP3_In(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal;

//clear interrupt flag:
    if (*USB_EP3R & 0x0080)  *USB_EP3R = *USB_EP3R & CLEAR_CTR_TX_MASK; //send ACK
//effectively there is nothing to notify
//fill the buffer = always send ZLP packets
    *gUSB_COUNT3_TX = 0;
//make EP TX "VALID":
    regVal = *USB_EP3R;
    if ((regVal & EPTX_DTOG2)==0) *USB_EP3R = 0x86A3; //EPTX_DTOG2=0x0020
    if ((regVal & EPTX_DTOG1)==0) *USB_EP3R = 0x8693; //EPTX_DTOG2=0x0010
}

// ###########################################################################
